/*
 * DESCRIPTION: kernel module for debugging registers and memory regions
 *              of mpc82xx via proc filesystem under entry-point
 *              /proc/driver/mpc82xx. Both read and write access are 
 *              supported.
 *
 * AUTHOR:      Alexander Hollinger (alexander.hollinger@gmx.net)
 *
 * DATE:        2003/07/04
 *
 * LICENSE:     This program is free software; you can redistribute it 
 *              and/or modify it under the terms of the GNU General Public 
 *		License as published by the Free Software Foundation; 
 *       	either version 2 of the License, or (at your option) any 
 *	        later version.
 *
 *	        This program is distributed in the hope that it will be 
 *      	useful, but WITHOUT ANY WARRANTY; without even the implied 
 *      	warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 *      	PURPOSE. See the GNU General Public License for more details.
 *
 *      	You should have received a copy of the GNU General Public 
 *      	License along with this program; if not, write to the Free 
 *      	Software Foundation, Inc., 59 Temple Place, Suite 330, 
 *      	Boston, MA  02111-1307 USA
 *
 * USAGE:       see README
 *
 */



#include <linux/module.h>
#include <linux/errno.h>
#include <linux/proc_fs.h>
#include <linux/init.h>
#include <linux/ctype.h>

#include <asm/mpc8260.h>

#ifndef CONFIG_PROC_FS
#error "CONFIG_PROC_FS must be enabled"
#endif






/* --------------------------- structures, variables, constants, declarations -------------------------- */


static struct proc_dir_entry *mpc82xx_proc_parent_dir;      /* handle to parent procfs dir               */
static char *mpc82xx_proc_node_register = "patchRegister";  /* name of procfs node to write to registers */
static char *mpc82xx_proc_node_memory   = "patchMemory";    /* name of procfs node to write to memory    */


typedef struct {                               /* information about one register or memory region        */
  char *name;                                  /* unique name of register/ memory region                 */
  unsigned int offset;                         /* offset from base address IMAP_ADDR/ or group-offset    */ 
  unsigned int size;                           /* size of register/ memory region in bytes               */
} mpc82xx_register_descr_t;


#define OFFSET_DIRECT   0                      /* direct access for groupoffset addess                   */
#define OFFSET_INDIRECT 1                      /* indirect access for groupoffset address                */

typedef struct mpc82xx_register_group{         /* information about a group of registers                 */
  char *groupname;                             /* description of group / name of procfs node             */          
  unsigned short groupoffset;                  /* group-offset from base address                         */
  mpc82xx_register_descr_t *reg_list;          /* pointer to 1st element of mpc82xx_register_descr_t[]   */
  int (*prt)(char*, int*, off_t*, off_t, int,
             struct mpc82xx_register_group*);  /* pointer to function that prints content of registers   */
  int groupoffset_address_mode;                /* flag for direct or indirect access (OFFSET_*)          */
} mpc82xx_register_group_t;


typedef struct {                               /* information about all groups of registers              */
  char *mapname;                               /* name of mpc82xx_memory_map / name of procfs dir        */
  mpc82xx_register_group_t *group_list;        /* pointer to 1st element of mpc82xx_register_group_t[]   */
} mpc82xx_memory_map_t;




static int mpc82xx_dump_register_group (char*, int*, off_t*, off_t, int, mpc82xx_register_group_t*);
static int mpc82xx_dump_memory_region  (char*, int*, off_t*, off_t, int, mpc82xx_register_group_t*);






/* --------------------------------------- definition of registers ------------------------------------- */



/* Instruction for adding a new register group
 *
 * 1. Create a new object of type mpc82xx_register_descr_t
 *    and initialize with array-elements of the 
 *    appropriate registers.
 * 2. Add an entry to mpc82xx_reg_groups with 
 *    name of register-group, pointer to 
 *    mpc82xx_register_descr object and pointer to
 *    function which handles output.
 */

 
static mpc82xx_register_descr_t reg_CPM_Dual_Port_RAM[] = {
  { "DPRAM1", 0x00000, 16384 },
  { "DPRAM2", 0x08000, 4096 },
  { "DPRAM3", 0x0b000, 4096 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_General_SIU[] = {
  { "SIUMCR",   0x10000, 4 },
  { "SYPCR",    0x10004, 4 },
  { "SWSR",     0x1000e, 2 },
  { "BCR",      0x10024, 4 },
  { "PPC_ACR",  0x10028, 1 },
  { "PPC_ALRH", 0x1002c, 4 },
  { "PPC_ALRL", 0x10030, 4 },
  { "LCL_ACR",  0x10034, 1 },
  { "LCL_ALRH", 0x10038, 4 },
  { "LCL_ALRL", 0x1003c, 4 },
  { "TESCR1",   0x10040, 4 },
  { "TESCR2",   0x10044, 4 },
  { "L_TESCR1", 0x10048, 4 },
  { "L_TESCR2", 0x1004c, 4 },
  { "PDTEA",    0x10050, 4 },
  { "PDTEM",    0x10054, 1 },
  { "LDTEA",    0x10058, 4 },
  { "LDTEM",    0x1005c, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_Memory_Controller[] = {
  { "BR0",   0x10100, 4 },
  { "OR0",   0x10104, 4 },
  { "BR1",   0x10108, 4 },
  { "OR1",   0x1010c, 4 },
  { "BR2",   0x10110, 4 },
  { "OR2",   0x10114, 4 },
  { "BR3",   0x10118, 4 },
  { "OR3",   0x1011c, 4 },
  { "BR4",   0x10120, 4 },
  { "OR4",   0x10124, 4 },
  { "BR5",   0x10128, 4 },
  { "OR5",   0x1012c, 4 },
  { "BR6",   0x10130, 4 },
  { "OR6",   0x10134, 4 },
  { "BR7",   0x10138, 4 },
  { "OR7",   0x1013c, 4 },
  { "BR8",   0x10140, 4 },
  { "OR8",   0x10144, 4 },
  { "BR9",   0x10148, 4 },
  { "OR9",   0x1014c, 4 },
  { "BR10",  0x10150, 4 },
  { "OR10",  0x10154, 4 },
  { "BR11",  0x10158, 4 },
  { "OR11",  0x1015c, 4 },
  { "MAR",   0x10168, 4 },
  { "MAMR",  0x10170, 4 },
  { "MBMR",  0x10174, 4 },
  { "MCMR",  0x10178, 4 },
  { "MPTPR", 0x10184, 2 },
  { "MDR",   0x10188, 4 },
  { "PSDMR", 0x10190, 4 },
  { "LSDMR", 0x10194, 4 },
  { "PURT",  0x10198, 1 },
  { "PSRT",  0x1019c, 1 },
  { "LURT",  0x101a0, 1 },
  { "LSRT",  0x101a4, 1 },
  { "IMMR",  0x101a8, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_System_Integration_Timers[] = {
  { "TMCNTSC", 0x10220, 2 },
  { "TMCNT",   0x10224, 4 },
  { "TMCNTAL", 0x1022c, 4 },
  { "PISCR",   0x10240, 2 },
  { "PITC",    0x10244, 4 },
  { "PITR",    0x10248, 4 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_Interrupt_Controller[] = {
  { "SICR",    0x10c00, 2 },
  { "SIVEC",   0x10c04, 4 },
  { "SIPNR_H", 0x10c08, 4 },
  { "SIPNR_L", 0x10c0c, 4 },
  { "SIPRR",   0x10c10, 4 },
  { "SCPRR_H", 0x10c14, 4 },
  { "SCPRR_L", 0x10c18, 4 },
  { "SIMR_H",  0x10c1c, 4 },
  { "SIMR_L",  0x10c20, 4 },
  { "SIEXR",   0x10c24, 4 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_Clocks_and_Reset[] = {
  { "SCCR", 0x10c80, 4 },
  { "SCMR", 0x10c88, 4 },
  { "RSR",  0x10c90, 4 },
  { "RMR",  0x10c94, 4 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_IO_Port[] = {
  { "PDIRA", 0x10d00, 4 },
  { "PPARA", 0x10d04, 4 },
  { "PSORA", 0x10d08, 4 },
  { "PODRA", 0x10d0c, 4 },
  { "PDATA", 0x10d10, 4 },
  { "PDIRB", 0x10d20, 4 },
  { "PPARB", 0x10d24, 4 },
  { "PSORB", 0x10d28, 4 },
  { "PODRB", 0x10d2c, 4 },
  { "PDATB", 0x10d30, 4 },
  { "PDIRC", 0x10d40, 4 },
  { "PPARC", 0x10d44, 4 },
  { "PSORC", 0x10d48, 4 },
  { "PODRC", 0x10d4c, 4 },
  { "PDATC", 0x10d50, 4 },
  { "PDIRD", 0x10d60, 4 },
  { "PPARD", 0x10d64, 4 },
  { "PSORD", 0x10d68, 4 },
  { "PODRD", 0x10d6c, 4 },
  { "PDATD", 0x10d70, 4 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_CPM_Timers[] = {
  { "TGCR1", 0x10d80, 1 },
  { "TGCR2", 0x10d84, 1 },
  { "TMR1",  0x10d90, 2 },
  { "TMR2",  0x10d92, 2 },
  { "TRR1",  0x10d94, 2 },
  { "TRR2",  0x10d96, 2 },
  { "TCR1",  0x10d98, 2 },
  { "TCR2",  0x10d9a, 2 },
  { "TCN1",  0x10d9c, 2 },
  { "TCN2",  0x10d9e, 2 },
  { "TMR3",  0x10da0, 2 },
  { "TMR4",  0x10da2, 2 },
  { "TRR3",  0x10da4, 2 },
  { "TRR4",  0x10da6, 2 },
  { "TCR3",  0x10da8, 2 },
  { "TCR4",  0x10daa, 2 },
  { "TCN3",  0x10dac, 2 },
  { "TCN4",  0x10dae, 2 },
  { "TER1",  0x10db0, 2 },
  { "TER2",  0x10db2, 2 },
  { "TER3",  0x10db4, 2 },
  { "TER4",  0x10db6, 2 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SDMA_General[] = {
  { "SDSR", 0x11018, 1 },
  { "SDMR", 0x1101c, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_IDMA[] = {
  { "IDSR1", 0x11020, 1 },
  { "IDMR1", 0x11024, 1 },
  { "IDSR2", 0x11028, 1 },
  { "IDMR2", 0x1102c, 1 },
  { "IDSR3", 0x11030, 1 },
  { "IDMR3", 0x11034, 1 },
  { "IDSR4", 0x11038, 1 },
  { "IDMR4", 0x1103c, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_FCC1[] = {
  { "GFMR1",       0x11300, 4 },
  { "FPSMR1",      0x11304, 4 },
  { "FTODR1",      0x11308, 2 },
  { "FDSR1",       0x1130c, 2 },
  { "FCCE1",       0x11310, 4 },
  { "FCCM1",       0x11314, 4 },
  { "FCCS1",       0x11318, 1 },
  { "FTIRR1_PHY0", 0x1131c, 1 },
  { "FTIRR1_PHY1", 0x1131d, 1 },
  { "FTIRR1_PHY2", 0x1131e, 1 },
  { "FTIRR1_PHY3", 0x1131f, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_FCC2[] = {
  { "GFMR2",       0x11320, 4 },
  { "FPSMR2",      0x11324, 4 },
  { "FTODR2",      0x11328, 2 },
  { "FDSR2",       0x1132c, 2 },
  { "FCCE2",       0x11330, 4 },
  { "FCCM2",       0x11334, 4 },
  { "FCCS2",       0x11338, 1 },
  { "FTIRR2_PHY0", 0x1133c, 1 },
  { "FTIRR2_PHY1", 0x1133d, 1 },
  { "FTIRR2_PHY2", 0x1133e, 1 },
  { "FTIRR2_PHY3", 0x1133f, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_FCC3[] = {
  { "GFMR3",  0x11340, 4 },
  { "FPSMR3", 0x11344, 4 },
  { "FTODR3", 0x11348, 2 },
  { "FDSR3",  0x1134c, 2 },
  { "FCCE3",  0x11350, 4 },
  { "FCCM3",  0x11354, 4 },
  { "FCCS3",  0x11358, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_BRG_5_8[] = {
  { "BRGC5", 0x115f0, 4 },
  { "BRGC6", 0x115f4, 4 },
  { "BRGC7", 0x115f8, 4 },
  { "BRGC8", 0x115fc, 4 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_I2C[] = {
  { "I2MOD", 0x11860, 1 },
  { "I2ADD", 0x11864, 1 },
  { "I2BRG", 0x11868, 1 },
  { "I2COM", 0x1186c, 1 },
  { "I2CER", 0x11870, 1 },
  { "I2CMR", 0x11874, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_Communications_Processor[] = {
  { "CPCR",  0x119c0, 4 },
  { "RCCR",  0x119c4, 4 },
  { "RTER",  0x119d6, 2 },
  { "RTMR",  0x119da, 2 },
  { "RTSCR", 0x119dc, 2 },
  { "RTSR",  0x119e0, 4 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_BRG_1_4[] = {
  { "BRGC1", 0x119f0, 4 },
  { "BRGC2", 0x119f4, 4 },
  { "BRGC3", 0x119f8, 4 },
  { "BRGC4", 0x119fc, 4 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SCC1[] = {
  { "GSMR_L1", 0x11a00, 4 },
  { "GSMR_H1", 0x11a04, 4 },
  { "PSMR1",   0x11a08, 2 },
  { "TODR1",   0x11a0c, 2 },
  { "DSR1",    0x11a0e, 2 },
  { "SCCE1",   0x11a10, 2 },
  { "SCCM1",   0x11a14, 2 },
  { "SCCS1",   0x11a17, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SCC2[] = {
  { "GSMR_L2", 0x11a20, 4 },
  { "GSMR_H2", 0x11a24, 4 },
  { "PSMR2",   0x11a28, 2 },
  { "TODR2",   0x11a2c, 2 },
  { "DSR2",    0x11a2e, 2 },
  { "SCCE2",   0x11a30, 2 },
  { "SCCM2",   0x11a34, 2 },
  { "SCCS2",   0x11a37, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SCC3[] = {
  { "GSMR_L3", 0x11a40, 4 },
  { "GSMR_H3", 0x11a44, 4 },
  { "PSMR3",   0x11a48, 2 },
  { "TODR3",   0x11a4c, 2 },
  { "DSR3",    0x11a4e, 2 },
  { "SCCE3",   0x11a50, 2 },
  { "SCCM3",   0x11a54, 2 },
  { "SCCS3",   0x11a57, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SCC4[] = {
  { "GSMR_L4", 0x11a60, 4 },
  { "GSMR_H4", 0x11a64, 4 },
  { "PSMR4",   0x11a68, 2 },
  { "TODR4",   0x11a6c, 2 },
  { "DSR4",    0x11a6e, 2 },
  { "SCCE4",   0x11a70, 2 },
  { "SCCM4",   0x11a74, 2 },
  { "SCCS4",   0x11a77, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SMC1[] = {
  { "SMCMR1", 0x11a82, 2 },
  { "SMCE1",  0x11a86, 1 },
  { "SMCM1",  0x11a8a, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SMC2[] = {
  { "SMCMR2", 0x11a92, 2 },
  { "SMCE2",  0x11a96, 1 },
  { "SMCM2",  0x11a9a, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SPI[] = {
  { "SPMODE", 0x11aa0, 2 },
  { "SPIE",   0x11aa6, 1 },
  { "SPIM",   0x11aaa, 1 },
  { "SPCOM",  0x11aad, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_CPM_Mux[] = {
  { "CMXSI1CR", 0x11b00, 1 },
  { "CMXSI2CR", 0x11b02, 1 },
  { "CMXFCR",   0x11b04, 4 },
  { "CMXSCR",   0x11b08, 4 },
  { "CMSSMR",   0x11b0c, 1 },
  { "CMXUAR",   0x11b0e, 2 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SI1_Registers[] = {
  { "SI1AMR",  0x11b20, 2 },
  { "SI1BMR",  0x11b22, 2 },
  { "SI1CMR",  0x11b24, 2 },
  { "SI1DMR",  0x11b26, 2 },
  { "SI1GMR",  0x11b28, 1 },
  { "SI1CMDR", 0x11b2a, 1 },
  { "SI1STR",  0x11b2c, 1 },
  { "SI1RSR",  0x11b2e, 2 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_MCC1_Registers[] = {
  { "MCCE1", 0x11b30, 2 },
  { "MCCM1", 0x11b34, 2 },
  { "MCCF1", 0x11b38, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SI2_Registers[] = {
  { "SI2AMR",  0x11b40, 2 },
  { "SI2BMR",  0x11b42, 2 },
  { "SI2CMR",  0x11b44, 2 },
  { "SI2DMR",  0x11b46, 2 },
  { "SI2GMR",  0x11b48, 1 },
  { "SI2CMDR", 0x11b4a, 1 },
  { "SI2STR",  0x11b4c, 1 },
  { "SI2RSR",  0x11b4e, 2 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_MCC2_Registers[] = {
  { "MCCE2", 0x11b50, 2 },
  { "MCCM2", 0x11b54, 2 },
  { "MCCF2", 0x11b58, 1 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SI1_RAM[] = {
  { "SI1TxRAM", 0x12000, 512 },
  { "SI1RxRAM", 0x12400, 512 },
  { NULL, 0, 0}
}; 


static mpc82xx_register_descr_t reg_SI2_RAM[] = {
  { "SI2TxRAM", 0x12800, 512 },
  { "SI2RxRAM", 0x12C00, 512 },
  { NULL, 0, 0}
}; 



/* ------------------------------------- definition of parameter RAM ----------------------------------- */


static mpc82xx_register_descr_t reg_SCC1_PRAM[] = {
  { "SCC1_PRAM_RBASE",  0x00, 2 },
  { "SCC1_PRAM_TBASE",  0x02, 2 },
  { "SCC1_PRAM_RFCR",   0x04, 1 },
  { "SCC1_PRAM_TFCR",   0x05, 1 },
  { "SCC1_PRAM_MRBLR",  0x06, 2 },
  { "SCC1_PRAM_RSTATE", 0x08, 4 },
  { "SCC1_PRAM_RIBP",   0x0c, 4 },
  { "SCC1_PRAM_RBPTR",  0x10, 2 },
  { "SCC1_PRAM_RIBC",   0x12, 2 },
  { "SCC1_PRAM_RTMP",   0x14, 4 },
  { "SCC1_PRAM_TSTATE", 0x18, 4 },
  { "SCC1_PRAM_TIBP",   0x1c, 4 },
  { "SCC1_PRAM_TBPTR",  0x20, 2 },
  { "SCC1_PRAM_TIBC",   0x22, 2 },
  { "SCC1_PRAM_TTMP",   0x24, 4 },
  { "SCC1_PRAM_RCRC",   0x28, 4 },
  { "SCC1_PRAM_TCRC",   0x2c, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_SCC2_PRAM[] = {
  { "SCC2_PRAM_RBASE",  0x00, 2 },
  { "SCC2_PRAM_TBASE",  0x02, 2 },
  { "SCC2_PRAM_RFCR",   0x04, 1 },
  { "SCC2_PRAM_TFCR",   0x05, 1 },
  { "SCC2_PRAM_MRBLR",  0x06, 2 },
  { "SCC2_PRAM_RSTATE", 0x08, 4 },
  { "SCC2_PRAM_RIBP",   0x0c, 4 },
  { "SCC2_PRAM_RBPTR",  0x10, 2 },
  { "SCC2_PRAM_RIBC",   0x12, 2 },
  { "SCC2_PRAM_RTMP",   0x14, 4 },
  { "SCC2_PRAM_TSTATE", 0x18, 4 },
  { "SCC2_PRAM_TIBP",   0x1c, 4 },
  { "SCC2_PRAM_TBPTR",  0x20, 2 },
  { "SCC2_PRAM_TIBC",   0x22, 2 },
  { "SCC2_PRAM_TTMP",   0x24, 4 },
  { "SCC2_PRAM_RCRC",   0x28, 4 },
  { "SCC2_PRAM_TCRC",   0x2c, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_SCC3_PRAM[] = {
  { "SCC3_PRAM_RBASE",  0x00, 2 },
  { "SCC3_PRAM_TBASE",  0x02, 2 },
  { "SCC3_PRAM_RFCR",   0x04, 1 },
  { "SCC3_PRAM_TFCR",   0x05, 1 },
  { "SCC3_PRAM_MRBLR",  0x06, 2 },
  { "SCC3_PRAM_RSTATE", 0x08, 4 },
  { "SCC3_PRAM_RIBP",   0x0c, 4 },
  { "SCC3_PRAM_RBPTR",  0x10, 2 },
  { "SCC3_PRAM_RIBC",   0x12, 2 },
  { "SCC3_PRAM_RTMP",   0x14, 4 },
  { "SCC3_PRAM_TSTATE", 0x18, 4 },
  { "SCC3_PRAM_TIBP",   0x1c, 4 },
  { "SCC3_PRAM_TBPTR",  0x20, 2 },
  { "SCC3_PRAM_TIBC",   0x22, 2 },
  { "SCC3_PRAM_TTMP",   0x24, 4 },
  { "SCC3_PRAM_RCRC",   0x28, 4 },
  { "SCC3_PRAM_TCRC",   0x2c, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_SCC4_PRAM[] = {
  { "SCC4_PRAM_RBASE",  0x00, 2 },
  { "SCC4_PRAM_TBASE",  0x02, 2 },
  { "SCC4_PRAM_RFCR",   0x04, 1 },
  { "SCC4_PRAM_TFCR",   0x05, 1 },
  { "SCC4_PRAM_MRBLR",  0x06, 2 },
  { "SCC4_PRAM_RSTATE", 0x08, 4 },
  { "SCC4_PRAM_RIBP",   0x0c, 4 },
  { "SCC4_PRAM_RBPTR",  0x10, 2 },
  { "SCC4_PRAM_RIBC",   0x12, 2 },
  { "SCC4_PRAM_RTMP",   0x14, 4 },
  { "SCC4_PRAM_TSTATE", 0x18, 4 },
  { "SCC4_PRAM_TIBP",   0x1c, 4 },
  { "SCC4_PRAM_TBPTR",  0x20, 2 },
  { "SCC4_PRAM_TIBC",   0x22, 2 },
  { "SCC4_PRAM_TTMP",   0x24, 4 },
  { "SCC4_PRAM_RCRC",   0x28, 4 },
  { "SCC4_PRAM_TCRC",   0x2c, 4 },
  { NULL, 0, 0}
};


static mpc82xx_register_descr_t reg_FCC1_PRAM[] = {
  { "FCC1_PRAM_RIPTR",   0x00, 2 },
  { "FCC1_PRAM_TIPTR",   0x02, 2 },
  { "FCC1_PRAM_MRBLR",   0x06, 2 },
  { "FCC1_PRAM_RSTATE",  0x08, 4 },
  { "FCC1_PRAM_RBASE",   0x0c, 4 },
  { "FCC1_PRAM_RBDSTAT", 0x10, 2 },
  { "FCC1_PRAM_RBDLEN",  0x12, 2 },
  { "FCC1_PRAM_RDPTR",   0x14, 4 },
  { "FCC1_PRAM_TSTATE",  0x18, 4 },
  { "FCC1_PRAM_TBASE",   0x1c, 4 },
  { "FCC1_PRAM_TBDSTAT", 0x20, 2 },
  { "FCC1_PRAM_TBDLEN",  0x22, 2 },
  { "FCC1_PRAM_TDPTR",   0x24, 4 },
  { "FCC1_PRAM_RBPTR",   0x28, 4 },
  { "FCC1_PRAM_TBPTR",   0x2c, 4 },
  { "FCC1_PRAM_RCRC",    0x30, 4 },
  { "FCC1_PRAM_TCRC",    0x34, 4 },
  { NULL, 0, 0}
};


static mpc82xx_register_descr_t reg_FCC2_PRAM[] = {
  { "FCC2_PRAM_RIPTR",   0x00, 2 },
  { "FCC2_PRAM_TIPTR",   0x02, 2 },
  { "FCC2_PRAM_MRBLR",   0x06, 2 },
  { "FCC2_PRAM_RSTATE",  0x08, 4 },
  { "FCC2_PRAM_RBASE",   0x0c, 4 },
  { "FCC2_PRAM_RBDSTAT", 0x10, 2 },
  { "FCC2_PRAM_RBDLEN",  0x12, 2 },
  { "FCC2_PRAM_RDPTR",   0x14, 4 },
  { "FCC2_PRAM_TSTATE",  0x18, 4 },
  { "FCC2_PRAM_TBASE",   0x1c, 4 },
  { "FCC2_PRAM_TBDSTAT", 0x20, 2 },
  { "FCC2_PRAM_TBDLEN",  0x22, 2 },
  { "FCC2_PRAM_TDPTR",   0x24, 4 },
  { "FCC2_PRAM_RBPTR",   0x28, 4 },
  { "FCC2_PRAM_TBPTR",   0x2c, 4 },
  { "FCC2_PRAM_RCRC",    0x30, 4 },
  { "FCC2_PRAM_TCRC",    0x34, 4 },
  { NULL, 0, 0}
};


static mpc82xx_register_descr_t reg_FCC3_PRAM[] = {
  { "FCC3_PRAM_RIPTR",   0x00, 2 },
  { "FCC3_PRAM_TIPTR",   0x02, 2 },
  { "FCC3_PRAM_MRBLR",   0x06, 2 },
  { "FCC3_PRAM_RSTATE",  0x08, 4 },
  { "FCC3_PRAM_RBASE",   0x0c, 4 },
  { "FCC3_PRAM_RBDSTAT", 0x10, 2 },
  { "FCC3_PRAM_RBDLEN",  0x12, 2 },
  { "FCC3_PRAM_RDPTR",   0x14, 4 },
  { "FCC3_PRAM_TSTATE",  0x18, 4 },
  { "FCC3_PRAM_TBASE",   0x1c, 4 },
  { "FCC3_PRAM_TBDSTAT", 0x20, 2 },
  { "FCC3_PRAM_TBDLEN",  0x22, 2 },
  { "FCC3_PRAM_TDPTR",   0x24, 4 },
  { "FCC3_PRAM_RBPTR",   0x28, 4 },
  { "FCC3_PRAM_TBPTR",   0x2c, 4 },
  { "FCC3_PRAM_RCRC",    0x30, 4 },
  { "FCC3_PRAM_TCRC",    0x34, 4 },
  { NULL, 0, 0}
};


static mpc82xx_register_descr_t reg_MCC1_PRAM[] = {
  { "MCC1_PRAM_MCCBASE",   0x00, 4 },
  { "MCC1_PRAM_MCCSTATE",  0x04, 2 },
  { "MCC1_PRAM_MRBLR",     0x06, 2 },
  { "MCC1_PRAM_GRFTHR",    0x08, 2 },
  { "MCC1_PRAM_GRFCNT",    0x0a, 2 },
  { "MCC1_PRAM_RINTTMP",   0x0c, 4 },
  { "MCC1_PRAM_DATA0",     0x10, 4 },
  { "MCC1_PRAM_DATA1",     0x14, 4 },
  { "MCC1_PRAM_TINTBASE",  0x18, 4 },
  { "MCC1_PRAM_TINTPTR",   0x1c, 4 },
  { "MCC1_PRAM_TINTTMP",   0x20, 4 },
  { "MCC1_PRAM_SCTPBASE",  0x24, 2 },
  { "MCC1_PRAM_C_MASK32",  0x28, 4 },
  { "MCC1_PRAM_XTRABASE",  0x2c, 2 },
  { "MCC1_PRAM_C_MASK16",  0x2e, 2 },
  { "MCC1_PRAM_RINTTMP0",  0x30, 4 },
  { "MCC1_PRAM_RINTTMP1",  0x34, 4 },
  { "MCC1_PRAM_RINTTMP2",  0x38, 4 },
  { "MCC1_PRAM_RINTTMP3",  0x3c, 4 },
  { "MCC1_PRAM_RINTBASE0", 0x40, 4 },
  { "MCC1_PRAM_RINTPTR0",  0x44, 4 },
  { "MCC1_PRAM_RINTBASE1", 0x48, 4 },
  { "MCC1_PRAM_RINTPTR1",  0x4c, 4 },
  { "MCC1_PRAM_RINTBASE2", 0x50, 4 },
  { "MCC1_PRAM_RINTPTR2",  0x54, 4 },
  { "MCC1_PRAM_RINTBASE3", 0x58, 4 },
  { "MCC1_PRAM_RINTPTR3",  0x5c, 4 },
  { "MCC1_PRAM_TS_TMP",    0x60, 4 },
  { NULL, 0, 0}
};


static mpc82xx_register_descr_t reg_MCC2_PRAM[] = {
  { "MCC2_PRAM_MCCBASE",   0x00, 4 },
  { "MCC2_PRAM_MCCSTATE",  0x04, 2 },
  { "MCC2_PRAM_MRBLR",     0x06, 2 },
  { "MCC2_PRAM_GRFTHR",    0x08, 2 },
  { "MCC2_PRAM_GRFCNT",    0x0a, 2 },
  { "MCC2_PRAM_RINTTMP",   0x0c, 4 },
  { "MCC2_PRAM_DATA0",     0x10, 4 },
  { "MCC2_PRAM_DATA1",     0x14, 4 },
  { "MCC2_PRAM_TINTBASE",  0x18, 4 },
  { "MCC2_PRAM_TINTPTR",   0x1c, 4 },
  { "MCC2_PRAM_TINTTMP",   0x20, 4 },
  { "MCC2_PRAM_SCTPBASE",  0x24, 2 },
  { "MCC2_PRAM_C_MASK32",  0x28, 4 },
  { "MCC2_PRAM_XTRABASE",  0x2c, 2 },
  { "MCC2_PRAM_C_MASK16",  0x2e, 2 },
  { "MCC2_PRAM_RINTTMP0",  0x30, 4 },
  { "MCC2_PRAM_RINTTMP1",  0x34, 4 },
  { "MCC2_PRAM_RINTTMP2",  0x38, 4 },
  { "MCC2_PRAM_RINTTMP3",  0x3c, 4 },
  { "MCC2_PRAM_RINTBASE0", 0x40, 4 },
  { "MCC2_PRAM_RINTPTR0",  0x44, 4 },
  { "MCC2_PRAM_RINTBASE1", 0x48, 4 },
  { "MCC2_PRAM_RINTPTR1",  0x4c, 4 },
  { "MCC2_PRAM_RINTBASE2", 0x50, 4 },
  { "MCC2_PRAM_RINTPTR2",  0x54, 4 },
  { "MCC2_PRAM_RINTBASE3", 0x58, 4 },
  { "MCC2_PRAM_RINTPTR3",  0x5c, 4 },
  { "MCC2_PRAM_TS_TMP",    0x60, 4 },
  { NULL, 0, 0}
};


static mpc82xx_register_descr_t reg_SMC1_PRAM[] = {
  { "SMC1_PRAM_RBASE",   0x00, 2 },
  { "SMC1_PRAM_TBASE",   0x02, 2 },
  { "SMC1_PRAM_RFCR",    0x04, 1 },
  { "SMC1_PRAM_TFCR",    0x05, 1 },
  { "SMC1_PRAM_MRBLR",   0x06, 2 },
  { "SMC1_PRAM_RSTATE",  0x08, 4 },
  { "SMC1_PRAM_RIDP",    0x0c, 4 },
  { "SMC1_PRAM_RBPTR",   0x10, 2 },
  { "SMC1_PRAM_RIBC",    0x12, 2 },
  { "SMC1_PRAM_RTMP",    0x14, 4 },
  { "SMC1_PRAM_TSTATE",  0x18, 4 },
  { "SMC1_PRAM_TIDP",    0x1c, 4 },
  { "SMC1_PRAM_TBPTR",   0x20, 2 },
  { "SMC1_PRAM_TIBC",    0x22, 2 },
  { "SMC1_PRAM_TTMP",    0x24, 4 },
  { "SMC1_PRAM_MAX_IDL", 0x28, 2 },
  { "SMC1_PRAM_IDLC",    0x2a, 2 },
  { "SMC1_PRAM_BRKLN",   0x2c, 2 },
  { "SMC1_PRAM_BRKEC",   0x2e, 2 },
  { "SMC1_PRAM_BRKCR",   0x30, 2 },
  { "SMC1_PRAM_R_MASK",  0x32, 2 },
  { "SMC1_PRAM_SDMATMP", 0x34, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_SMC2_PRAM[] = {
  { "SMC2_PRAM_RBASE",   0x00, 2 },
  { "SMC2_PRAM_TBASE",   0x02, 2 },
  { "SMC2_PRAM_RFCR",    0x04, 1 },
  { "SMC2_PRAM_TFCR",    0x05, 1 },
  { "SMC2_PRAM_MRBLR",   0x06, 2 },
  { "SMC2_PRAM_RSTATE",  0x08, 4 },
  { "SMC2_PRAM_RIDP",    0x0c, 4 },
  { "SMC2_PRAM_RBPTR",   0x10, 2 },
  { "SMC2_PRAM_RIBC",    0x12, 2 },
  { "SMC2_PRAM_RTMP",    0x14, 4 },
  { "SMC2_PRAM_TSTATE",  0x18, 4 },
  { "SMC2_PRAM_TIDP",    0x1c, 4 },
  { "SMC2_PRAM_TBPTR",   0x20, 2 },
  { "SMC2_PRAM_TIBC",    0x22, 2 },
  { "SMC2_PRAM_TTMP",    0x24, 4 },
  { "SMC2_PRAM_MAX_IDL", 0x28, 2 },
  { "SMC2_PRAM_IDLC",    0x2a, 2 },
  { "SMC2_PRAM_BRKLN",   0x2c, 2 },
  { "SMC2_PRAM_BRKEC",   0x2e, 2 },
  { "SMC2_PRAM_BRKCR",   0x30, 2 },
  { "SMC2_PRAM_R_MASK",  0x32, 2 },
  { "SMC2_PRAM_SDMATMP", 0x34, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_IDMA1_PRAM[] = {
  { "IDMA1_PRAM_IBASE",       0x00, 2 },
  { "IDMA1_PRAM_DMC",         0x02, 2 },
  { "IDMA1_PRAM_IBDPTR",      0x04, 2 },
  { "IDMA1_PRAM_DPR_BUF",     0x06, 2 },
  { "IDMA1_PRAM_BUF_INV",     0x08, 2 },
  { "IDMA1_PRAM_SS_MAX",      0x0a, 2 },
  { "IDMA1_PRAM_DPR_IN_PTR",  0x0c, 2 },
  { "IDMA1_PRAM_STS",         0x0e, 2 },
  { "IDMA1_PRAM_DPR_OUT_PTR", 0x10, 2 },
  { "IDMA1_PRAM_SEOB",        0x12, 2 },
  { "IDMA1_PRAM_DEOB",        0x14, 2 },
  { "IDMA1_PRAM_DTS",         0x16, 2 },
  { "IDMA1_PRAM_RET_ADD",     0x18, 2 },
  { "IDMA1_PRAM_BD_CNT",      0x1c, 4 },
  { "IDMA1_PRAM_S_PTR",       0x20, 4 },
  { "IDMA1_PRAM_D_PTR",       0x24, 4 },
  { "IDMA1_PRAM_ISTATE",      0x28, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_IDMA2_PRAM[] = {
  { "IDMA2_PRAM_IBASE",       0x00, 2 },
  { "IDMA2_PRAM_DMC",         0x02, 2 },
  { "IDMA2_PRAM_IBDPTR",      0x04, 2 },
  { "IDMA2_PRAM_DPR_BUF",     0x06, 2 },
  { "IDMA2_PRAM_BUF_INV",     0x08, 2 },
  { "IDMA2_PRAM_SS_MAX",      0x0a, 2 },
  { "IDMA2_PRAM_DPR_IN_PTR",  0x0c, 2 },
  { "IDMA2_PRAM_STS",         0x0e, 2 },
  { "IDMA2_PRAM_DPR_OUT_PTR", 0x10, 2 },
  { "IDMA2_PRAM_SEOB",        0x12, 2 },
  { "IDMA2_PRAM_DEOB",        0x14, 2 },
  { "IDMA2_PRAM_DTS",         0x16, 2 },
  { "IDMA2_PRAM_RET_ADD",     0x18, 2 },
  { "IDMA2_PRAM_BD_CNT",      0x1c, 4 },
  { "IDMA2_PRAM_S_PTR",       0x20, 4 },
  { "IDMA2_PRAM_D_PTR",       0x24, 4 },
  { "IDMA2_PRAM_ISTATE",      0x28, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_IDMA3_PRAM[] = {
  { "IDMA3_PRAM_IBASE",       0x00, 2 },
  { "IDMA3_PRAM_DMC",         0x02, 2 },
  { "IDMA3_PRAM_IBDPTR",      0x04, 2 },
  { "IDMA3_PRAM_DPR_BUF",     0x06, 2 },
  { "IDMA3_PRAM_BUF_INV",     0x08, 2 },
  { "IDMA3_PRAM_SS_MAX",      0x0a, 2 },
  { "IDMA3_PRAM_DPR_IN_PTR",  0x0c, 2 },
  { "IDMA3_PRAM_STS",         0x0e, 2 },
  { "IDMA3_PRAM_DPR_OUT_PTR", 0x10, 2 },
  { "IDMA3_PRAM_SEOB",        0x12, 2 },
  { "IDMA3_PRAM_DEOB",        0x14, 2 },
  { "IDMA3_PRAM_DTS",         0x16, 2 },
  { "IDMA3_PRAM_RET_ADD",     0x18, 2 },
  { "IDMA3_PRAM_BD_CNT",      0x1c, 4 },
  { "IDMA3_PRAM_S_PTR",       0x20, 4 },
  { "IDMA3_PRAM_D_PTR",       0x24, 4 },
  { "IDMA3_PRAM_ISTATE",      0x28, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_IDMA4_PRAM[] = {
  { "IDMA4_PRAM_IBASE",       0x00, 2 },
  { "IDMA4_PRAM_DMC",         0x02, 2 },
  { "IDMA4_PRAM_IBDPTR",      0x04, 2 },
  { "IDMA4_PRAM_DPR_BUF",     0x06, 2 },
  { "IDMA4_PRAM_BUF_INV",     0x08, 2 },
  { "IDMA4_PRAM_SS_MAX",      0x0a, 2 },
  { "IDMA4_PRAM_DPR_IN_PTR",  0x0c, 2 },
  { "IDMA4_PRAM_STS",         0x0e, 2 },
  { "IDMA4_PRAM_DPR_OUT_PTR", 0x10, 2 },
  { "IDMA4_PRAM_SEOB",        0x12, 2 },
  { "IDMA4_PRAM_DEOB",        0x14, 2 },
  { "IDMA4_PRAM_DTS",         0x16, 2 },
  { "IDMA4_PRAM_RET_ADD",     0x18, 2 },
  { "IDMA4_PRAM_BD_CNT",      0x1c, 4 },
  { "IDMA4_PRAM_S_PTR",       0x20, 4 },
  { "IDMA4_PRAM_D_PTR",       0x24, 4 },
  { "IDMA4_PRAM_ISTATE",      0x28, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_SPI_PRAM[] = {
  { "SPI_PRAM_RBASE",   0x00, 2 },
  { "SPI_PRAM_TBASE",   0x02, 2 },
  { "SPI_PRAM_RFCR",    0x04, 1 },
  { "SPI_PRAM_TFCR",    0x05, 1 },
  { "SPI_PRAM_MRBLR",   0x06, 2 },
  { "SPI_PRAM_RSTATE",  0x08, 4 },
  { "SPI_PRAM_RIDP",    0x0c, 4 },
  { "SPI_PRAM_RBPTR",   0x10, 2 },
  { "SPI_PRAM_RIBC",    0x12, 2 },
  { "SPI_PRAM_RTMP",    0x14, 4 },
  { "SPI_PRAM_TSTATE",  0x18, 4 },
  { "SPI_PRAM_TIDP",    0x1c, 4 },
  { "SPI_PRAM_TBPTR",   0x20, 2 },
  { "SPI_PRAM_TIBC",    0x22, 2 },
  { "SPI_PRAM_TTMP",    0x24, 4 },
  { "SPI_PRAM_SDMATMP", 0x34, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_RISC_Timers_PRAM[] = {
  { "RISC_Timers_PRAM_TM_BASE", 0x00, 2 },
  { "RISC_Timers_PRAM_TM_PTR",  0x02, 2 },
  { "RISC_Timers_PRAM_R_TMR",   0x04, 2 },
  { "RISC_Timers_PRAM_R_TMV",   0x06, 2 },
  { "RISC_Timers_PRAM_TM_CMD",  0x08, 4 },
  { "RISC_Timers_PRAM_TM_CNT",  0x0c, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_descr_t reg_I2C_PRAM[] = {
  { "I2C_PRAM_RBASE",  0x00, 2 },
  { "I2C_PRAM_TBASE",  0x02, 2 },
  { "I2C_PRAM_RFCR",   0x04, 1 },
  { "I2C_PRAM_TFCR",   0x05, 1 },
  { "I2C_PRAM_MRBLR",  0x06, 2 },
  { "I2C_PRAM_RSTATE", 0x08, 4 },
  { "I2C_PRAM_RPTR",   0x0c, 4 },
  { "I2C_PRAM_RBPTR",  0x10, 2 },
  { "I2C_PRAM_RCOUNT", 0x12, 2 },
  { "I2C_PRAM_RTEMP",  0x14, 4 },
  { "I2C_PRAM_TSTATE", 0x18, 4 },
  { "I2C_PRAM_TPTR",   0x1c, 4 },
  { "I2C_PRAM_TBPTR",  0x20, 2 },
  { "I2C_PRAM_TCOUNT", 0x22, 2 },
  { "I2C_PRAM_TTEMP",  0x24, 4 },
  { NULL, 0, 0}
};  


static mpc82xx_register_group_t mpc82xx_reg_groups[] = {
  { "cpm_dual_port_ram",    0x0000, reg_CPM_Dual_Port_RAM,         mpc82xx_dump_memory_region,  OFFSET_DIRECT   },
  { "general_siu",          0x0000, reg_General_SIU,               mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "mem_ctrl",             0x0000, reg_Memory_Controller,         mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "sys_intgr_timers",     0x0000, reg_System_Integration_Timers, mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "interrupt_controller", 0x0000, reg_Interrupt_Controller,      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "clocks_reset",         0x0000, reg_Clocks_and_Reset,          mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "io_port",              0x0000, reg_IO_Port,                   mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "cpm_timers",           0x0000, reg_CPM_Timers,                mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "sdma_general",         0x0000, reg_SDMA_General,              mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "idma",                 0x0000, reg_IDMA,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "fcc1",                 0x0000, reg_FCC1,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "fcc2",                 0x0000, reg_FCC2,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "fcc3",                 0x0000, reg_FCC3,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "brg_5_8",              0x0000, reg_BRG_5_8,                   mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "i2c",                  0x0000, reg_I2C,                       mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "comm_proc",            0x0000, reg_Communications_Processor,  mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "brg_1_4",              0x0000, reg_BRG_1_4,                   mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "scc1",                 0x0000, reg_SCC1,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "scc2",                 0x0000, reg_SCC2,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "scc3",                 0x0000, reg_SCC3,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "scc4",                 0x0000, reg_SCC4,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "smc1",                 0x0000, reg_SMC1,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "smc2",                 0x0000, reg_SMC2,                      mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "spi",                  0x0000, reg_SPI,                       mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "cpm_mux",              0x0000, reg_CPM_Mux,                   mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "si1_registers",        0x0000, reg_SI1_Registers,             mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "mcc1_registers",       0x0000, reg_MCC1_Registers,            mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "si2_registers",        0x0000, reg_SI2_Registers,             mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "mcc2_registers",       0x0000, reg_MCC2_Registers,            mpc82xx_dump_register_group, OFFSET_DIRECT   }, 
  { "si1_ram",              0x0000, reg_SI1_RAM,                   mpc82xx_dump_memory_region,  OFFSET_DIRECT   },
  { "si2_ram",              0x0000, reg_SI2_RAM,                   mpc82xx_dump_memory_region,  OFFSET_DIRECT   },

  { "scc1_pram",            0x8000, reg_SCC1_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "scc2_pram",            0x8100, reg_SCC2_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "scc3_pram",            0x8200, reg_SCC3_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "scc4_pram",            0x8300, reg_SCC4_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "fcc1_pram",            0x8400, reg_FCC1_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "fcc2_pram",            0x8500, reg_FCC2_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "fcc3_pram",            0x8600, reg_FCC3_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "mcc1_pram",            0x8700, reg_MCC1_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },
  { "mcc2_pram",            0x8800, reg_MCC2_PRAM,                 mpc82xx_dump_register_group, OFFSET_DIRECT   },

  { "smc1_pram",            0x87fc, reg_SMC1_PRAM,                 mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "smc2_pram",            0x88fc, reg_SMC2_PRAM,                 mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "idma1_pram",           0x87fe, reg_IDMA1_PRAM,                mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "idma2_pram",           0x88fe, reg_IDMA2_PRAM,                mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "idma3_pram",           0x89fe, reg_IDMA3_PRAM,                mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "idma4_pram",           0x8afe, reg_IDMA4_PRAM,                mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "spi_pram",             0x89fc, reg_SPI_PRAM,                  mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "risc_timers_pram",     0x8ae0, reg_RISC_Timers_PRAM,          mpc82xx_dump_register_group, OFFSET_INDIRECT },
  { "i2c_pram",             0x8afc, reg_I2C_PRAM,                  mpc82xx_dump_register_group, OFFSET_INDIRECT },

  { NULL, 0, NULL, NULL, 0}
};


static mpc82xx_memory_map_t mpc82xx_mem_map = {"mpc82xx", mpc82xx_reg_groups};




/* ------------------------------ print content of registers ------------------------ */



/*  PRINT_PROC
 *  This macro frees the machine specific function from bounds checking and
 *  this like that... 
 */
#define PRINT_PROC(fmt,args...) \
        do { \
                *len += sprintf( buffer+*len, fmt, ##args ); \
                if (*begin + *len > offset + size) \
                        return( 0 ); \
                if (*begin + *len < offset) { \
                        *begin += *len; \
                        *len = 0; \
                } \
        } while(0)



/**
 *  print content of registers which are part of the register group 'reg_group' to procfs
 *  (IMAP_ADDR + reg_group->groupoffset) _is_ baseaddress
 *  display content in formatted list
 */
static int mpc82xx_dump_register_group (char *buffer, int *len, off_t *begin, off_t offset, int size, mpc82xx_register_group_t *reg_group)
{
  
  mpc82xx_register_descr_t *reg;
  unsigned short groupoffset;


  /* get groupoffset according to specified addressing mode */
  if ((reg_group->groupoffset_address_mode) == OFFSET_DIRECT) {
    groupoffset = reg_group->groupoffset;                                     /* groupoffset */
  } else {
    groupoffset = *((unsigned short*)(IMAP_ADDR + reg_group->groupoffset));   /* pointer to groupoffset */
  }


  /* print groupname and header for listing */
  PRINT_PROC("register group: %s, group offset: %#06x\n", reg_group->groupname, groupoffset);
  PRINT_PROC("  register                   offset       size  value\n");


  /* print content of all registers in this group */
  reg = reg_group->reg_list;        
  for (reg = reg_group->reg_list; reg->name != NULL; reg++) {

    /* print info about register */
    PRINT_PROC("   %-25s  %#-10x   %u     ", reg->name,  reg->offset, reg->size);
    switch(reg->size){
      case 1: 
	PRINT_PROC("%02x\n", *((unsigned char*)(IMAP_ADDR + groupoffset + reg->offset)) );
	break;
      case 2: 
	PRINT_PROC("%04x\n", *((unsigned short*)(IMAP_ADDR + groupoffset + reg->offset)) );
	break;
      case 4: 
	PRINT_PROC("%08x\n", *((unsigned int*)(IMAP_ADDR + groupoffset + reg->offset)) );
	break;
      default: 
	PRINT_PROC("???\n");
	break;
    } /* switch */

  } /* for */

  return 1;        /* end of output */


} /* mpc82xx_dump_register_group */



/**
 *  print content of memory regions which are part of the register group 'reg_group' to procfs 
 *  display content as memory dump
 *  (always OFFSET_DIRECT)
 */
static int mpc82xx_dump_memory_region(char *buffer, int *len, off_t *begin, off_t offset, int size, mpc82xx_register_group_t *reg_group)
{
  
  mpc82xx_register_descr_t *reg;
  int i, n;

   
  /* print groupname for listing */
  PRINT_PROC("memory group: %s, group offset: %#06x\n\n", reg_group->groupname, reg_group->groupoffset);

  /* print memory content of all memory regions in this group */
  for (reg = reg_group->reg_list; reg->name != NULL; reg++) {
  

    /* print register information */
    PRINT_PROC("  dump of memory region %s, offset %#x, size %i bytes\n",
	       reg->name, reg->offset, reg->size);


    /* content of one line of output */
    for (i=0; i<reg->size; i=i+16){
      
      /* print offset */
      PRINT_PROC("    %#08x  ", reg->offset + i); 

      /* print hexadecimal value */
      for (n=0; n<16; n=n+4){
        if (i+n < reg->size) {
          PRINT_PROC("%08x ", *((unsigned int*)(IMAP_ADDR + reg_group->groupoffset + reg->offset + i + n)) );
	}
      }
 
      PRINT_PROC("\n");

    } /* for */

    PRINT_PROC("\n");
	 
  } /* for */

  return 1;        /* end of output */


} /* mpc82xx_dump_memory_region */



/**
 *  search for register in mpc82xx_mem_map.group_list
 *  arguments:
 *    *name          pointer to the string where the name is found
 *    size           only the first size bytes of *name are part of the register name
 *    len_value      length of input value
 *    **reg_found    pointer to register if search was successful (return_value==1)
 *    **group_found  pointer to group if search was successful (return_value==1)
 *  return value:
 *    0                   register not found
 *    1                   register found and valid
 */
static int mpc82xx_is_valid_register(const char* name, const size_t size, const int len_value, mpc82xx_register_descr_t **reg_found, mpc82xx_register_group_t **group_found)
{
    mpc82xx_register_group_t *reg_group;
    mpc82xx_register_descr_t *reg;  
    int return_value = 0;
    int found_flag = 0;
 


    /* search for register */
    for (reg_group = mpc82xx_mem_map.group_list; reg_group->groupname != NULL; reg_group++) {
      for (reg = reg_group->reg_list; reg->name != NULL; reg++) {
	if ( strncmp(reg->name, name, size) == 0 ) {

	  /* check if input of value has correct size. */
	  if ( len_value <= (reg->size * 2) ) {
	    return_value = 1;         /* search was successful   */
	    *reg_found = reg;          /* get pointer to register */
	    *group_found = reg_group;  /* get pointer to group    */
	  }

	found_flag = 1;             /* register found          */ 
	break;
	  
        } /* if */
      } /* for */
      
      if (found_flag == 1) { break; } /* register found, break loop */
      
    } /* for */
    
    if (found_flag == 0) { 
      return_value = 0;               /* register was not found */
    }

    return return_value;   /* 1 if register found and correct, -EINVAL if not */

}



/**
 *  change content of a specified register/ memory cell
 *  register: (has to be part of register group mpc82xx_mem_map)
 *    usage:   echo "<register_name>=0x<hexvalue>" > /proc/driver/mpc82xx/patchRegister
 *    example: echo "GSMR_L1=0x90abcdef" > /proc/driver/mpc82xx/patchRegister
 *  memory:
 *    usage:   echo "<address>=0x<hexvalue>" > /proc/driver/mpc82xx/patchMemory
 *    example: echo "012df0=0x12345678" > /proc/driver/mpc82xx/patchMemory
 */
static int mpc82xx_write_proc (struct file *file, const char *buffer, unsigned long count, void *data)
{
  char *sep; 
  char *hex;
  unsigned int value = 0;
  int len_value;


  /* check for separator '=' */
  if ( (sep = strchr(buffer, (int)('=')) ) == NULL) { 
    return -EINVAL;
  }
  

  /* check for hex value */
  if ( (hex = strchr(sep+1, (int)('x')) ) == NULL) { 
    return -EINVAL;
  }


  /* get value from buffer - convert from hexadecimal (16) */
  value = (int) simple_strtol( hex+1, (char**)(buffer+count-1), 16 );
  len_value = (buffer+count-1) - (hex+1);

  
  /* check if we do register patching or memory patching */
  if ( data == mpc82xx_proc_node_register ) {
    
    /* patch register */

    mpc82xx_register_descr_t *reg = NULL; 
    mpc82xx_register_group_t *reg_group = NULL;
    unsigned short groupoffset;

    /* check if input of registername is a correct register and length of value is of valid size */
    if (mpc82xx_is_valid_register(buffer, (size_t)(sep - buffer), len_value, &reg, &reg_group) == 1) {


      /* get groupoffset according to specified addressing mode */
      if (reg_group->groupoffset_address_mode == OFFSET_DIRECT) {
	groupoffset = reg_group->groupoffset;                                     /* groupoffset */
      } else {
      	groupoffset = *((unsigned short*)(IMAP_ADDR + reg_group->groupoffset));   /* pointer to groupoffset */
      }
      
      /* change value */
      switch(reg->size){
        case 1:
	  *(unsigned char*)(IMAP_ADDR + groupoffset + reg->offset) = (unsigned char)value;
	  break;
        case 2:
	  *(unsigned short*)(IMAP_ADDR + groupoffset + reg->offset) = (unsigned short)value;
	  break;
        case 4:
	  *(unsigned int*)(IMAP_ADDR + groupoffset + reg->offset) = (unsigned int)value;
	  break;
      } /* switch */
	    
    } else {
      return -EINVAL;  /* error occured */
    }
	  

  } else {


    /* patch memory */
    
    /* argument is address */
    unsigned int addressoffset;
    addressoffset = (unsigned int) simple_strtol( buffer, (char**)sep, 16 );    

    /* only allow values up to one word; if less filling up with zero from left by default */
    if (len_value <= 8) {
      *(unsigned int*)(IMAP_ADDR + addressoffset) = (unsigned int)value;   /* patch Memory */
    } else {
      return -EINVAL;                 /* length of value of input is too long */
    }


  }
  

  return count;


} /* mpc82xx_write_proc */



/**
 *  print usage of write-access to registers
 */
static int mpc82xx_dump_write_info (char *buffer, int *len, off_t *begin, off_t offset, int size, void *arg)
{
  
  if ( arg == mpc82xx_proc_node_register ) {
    PRINT_PROC("patchRegister: change content of the specified register\n");
    PRINT_PROC("  usage:   <register_name>=0x<hexvalue>\n");
    PRINT_PROC("           <register_name> := name of register\n");
    PRINT_PROC("           <hexvalue>      := new hexadecimal value of register\n");
    PRINT_PROC("  example: echo 'GSMR_L1=0x90abcdef' > /proc/driver/mpc82xx/patchRegister\n");
  } else {
    PRINT_PROC("patchMemory: writes a word to the specified memory cell\n");
    PRINT_PROC("  usage:   <address>=0x<hexvalue>\n");
    PRINT_PROC("           <address>  := offset from IMAP_ADDR\n");
    PRINT_PROC("           <hexvalue> := new hexadecimal value of memory cell\n");
    PRINT_PROC("  example: echo '012df0=0x12345678' > /proc/driver/mpc82xx/patchMemory\n");
  }

  return 1;        /* end of output */


} /* mpc82xx_dump_write_info */



static int mpc82xx_write_proc_info (char *buffer, char **start, off_t offset, int size, int *eof, void *arg)
{
  int len = 0;
  off_t begin = 0;

  *eof = mpc82xx_dump_write_info( buffer, &len, &begin, offset, size, arg );

  if (offset >= begin + len)  return 0;

  *start = buffer + (offset - begin);
  return size < begin + len - offset ? size : begin + len - offset;
}


/* --------------------------- PROC FS entry / exit points -------------------------- */




/**
 *  read operation was applied on PROC FS entry
 */
static int mpc82xx_read_proc (char *buffer, char **start, off_t offset, int size, int *eof, void *arg)
{

  int   len = 0;
  off_t begin = 0;
  mpc82xx_register_group_t *reg_group;  

  reg_group = (mpc82xx_register_group_t*) arg;     /* arg is of type mpc82xx_register_group_t */
  *eof = ( *(reg_group)->prt ) (buffer, &len, &begin, offset, size, reg_group);

  if (offset >= begin + len)  return 0;

  *start = buffer + (offset - begin);
  return size < begin + len - offset ? size : begin + len - offset;


} /* mpc82xx_read_proc */



/**
 *  init entry point
 */
static int __init mpc82xx_proc_init (void)
{

  mpc82xx_register_group_t *group;
  struct proc_dir_entry *mpc82xx_proc_entry;


  printk("loading module mpc82xx_proc...");   /* print status */

  /* create procfs directory */
  mpc82xx_proc_parent_dir = proc_mkdir(mpc82xx_mem_map.mapname, proc_root_driver);
  if(!mpc82xx_proc_parent_dir){
    printk(KERN_ERR "\n%s (%d): cannot create directory %s \n", __FILE__,__LINE__, mpc82xx_mem_map.mapname);    
    return -EINVAL;                                        /* creating procfs directory failed */
  }  

  /* create/link procfs nodes for each group (read-only) */
  for (group = mpc82xx_mem_map.group_list; group->groupname != NULL; group++) {
    if(!create_proc_read_entry(group->groupname, 0, mpc82xx_proc_parent_dir, mpc82xx_read_proc, group)){
      printk(KERN_ERR "\n%s (%d): cannot create %s \n", __FILE__,__LINE__, group->groupname);
      return -EINVAL;                                      /* creating procfs nodes failed */
    }
  }

  /* create node entries for write access */
  if( (mpc82xx_proc_entry = create_proc_entry(mpc82xx_proc_node_register, 0, mpc82xx_proc_parent_dir)) == NULL){
    printk(KERN_ERR "\n%s (%d): cannot create %s \n", __FILE__,__LINE__, mpc82xx_proc_node_register);
    return -EINVAL;                                        /* creating procfs nodes failed */
  } else {
    mpc82xx_proc_entry->read_proc  = mpc82xx_write_proc_info;
    mpc82xx_proc_entry->write_proc = mpc82xx_write_proc;
    mpc82xx_proc_entry->data = mpc82xx_proc_node_register;
  }
  if( (mpc82xx_proc_entry = create_proc_entry(mpc82xx_proc_node_memory, 0, mpc82xx_proc_parent_dir)) == NULL){
    printk(KERN_ERR "\n%s (%d): cannot create %s \n", __FILE__,__LINE__, mpc82xx_proc_node_memory);
    return -EINVAL;                                        /* creating procfs nodes failed */
  } else {
    mpc82xx_proc_entry->read_proc  = mpc82xx_write_proc_info;
    mpc82xx_proc_entry->write_proc = mpc82xx_write_proc;
    mpc82xx_proc_entry->data = mpc82xx_proc_node_memory;
  }
  

 
  printk("   [OK]\n"); 
  return 0;    /* success */


} /* mpc82xx_proc_init */



/**
 *  cleanup entry point
 */ 
static void __exit mpc82xx_proc_cleanup (void)
{ 

  mpc82xx_register_group_t *group; 

  printk("removing module mpc82xx_proc...");    /* print status */

  /* remove procfs-entries */
  remove_proc_entry(mpc82xx_proc_node_register, mpc82xx_proc_parent_dir); /* remove write nodes */
  remove_proc_entry(mpc82xx_proc_node_memory,   mpc82xx_proc_parent_dir); /* remove write nodes */

  for (group = mpc82xx_mem_map.group_list; group->groupname != NULL; group++) {
    remove_proc_entry(group->groupname, mpc82xx_proc_parent_dir);         /* remove read nodes */
  }
  remove_proc_entry(mpc82xx_mem_map.mapname, proc_root_driver);           /* remove dir */

  printk("   [OK]\n");

} /* mpc82xx_proc_cleanup */



/**
 *  for loading the module dynamically
 */
module_init(mpc82xx_proc_init);
module_exit(mpc82xx_proc_cleanup);



/**
 *  about driver
 */
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Alexander Hollinger <alexander.hollinger@gmx.net>");
MODULE_DESCRIPTION("PROC FS entries for Internal Memory Map of MPC82xx");
